
    (*********************************************************************
     *  AlphaBlend.pas                                                   *
     *                                                                   *
     *  author    : Milenko Mitrovic                                     *
     *  email     : dcoder@dsp-worx.de                                   *
     *  web       : http://dsp-worx.de                                   *
     *  date      : 24-07-2003                                           *
     *                                                                   *
     *  The contents of this file are used with permission, subject to   *
     *  the Mozilla Public License Version 1.1 (the "License"); you may  *
     *  not use this file except in compliance with the License. You may *
     *  obtain a copy of the License at                                  *
     *  http://www.mozilla.org/MPL/MPL-1.1.html                          *
     *                                                                   *
     *  Software distributed under the License is distributed on an      *
     *  "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or   *
     *  implied. See the License for the specific language governing     *
     *  rights and limitations under the License.                        *
     *                                                                   *
     *  Blendit_ASM is taken from TBSMorphButton v1.0 (C) Babak Sateli   *
     *  MMX optimized Blendit_MMX (C) Milenko Mitrovic                   *
     *  MMX Version works more then 2x faster then the ASM Version :)    *
     *                                                                   *
     *  Bitmaps must be 24 Bit, Dens must be in a range of 0..64         *
     *                                                                   *
     *  (C) 2003, 2004 Milenko Mitrovic <dcoder@dsp-worx.de>             *
     *                                                                   *
     *********************************************************************)

unit AlphaBlend;

interface

uses
  Graphics;

  procedure Blendit(bFr,bTo,bLn : TBitmap; Dens : LongInt);

implementation

const
  Mask0101 = $00FF00FF;
  Mask1010 = $FF00FF00;

var
  ESI, EDI, ESP, EBP, EBX,
  FinA,
  Dens1, Dens2 : Longint;
  FBlend : procedure (bFr,bTo,bLn : Pointer; Width,Height : integer; Dens : LongInt);

procedure Blendit_ASM(bFr,bTo,bLn : Pointer ; Width,Height : Integer ; Dens : LongInt);
asm
        mov       &ebx, ebx
        mov       &edi, edi
        mov       &esi, esi
        mov       &esp, esp
        mov       &ebp, ebp
        mov       ebx, Dens
        mov       Dens1, ebx
        neg       bl
        add       bl, 64
        mov       Dens2, ebx
        cmp       Dens1, 0
        jz        @END
        mov       edi, bFr
        mov       esi, bTo
        mov       ecx, bLn
        mov       eax, Width
        lea       eax, [eax+eax*2+3]
        and       eax, $FFFFFFFC
        imul      Height
        add       eax, edi
        mov       FinA, eax
        mov       ebp,edi
        mov       esp,esi
        mov       ecx,ecx
@LOOP:
        mov       eax, [ebp]
        mov       edi, [esp]
        mov       ebx, eax
        and       eax, Mask1010
        and       ebx, Mask0101
        shr       eax, 6
        imul      eax, Dens2
        imul      ebx, Dens2
        mov       esi, edi
        and       edi, Mask1010
        and       esi, Mask0101
        shr       edi, 6
        imul      edi, Dens1
        imul      esi, Dens1
        add       eax, edi
        add       ebx, esi
        and       eax, Mask1010
        shr       ebx, 6
        and       ebx, Mask0101
        or        eax, ebx
        mov       [ecx], eax
        add       ebp, 4
        add       esp, 4
        add       ecx, 4
        cmp       ebp, FinA
        jne       @LOOP
@END:
        mov       ebx, &ebx
        mov       edi, &edi
        mov       esi, &esi
        mov       esp, &esp
        mov       ebp, &ebp
end;

procedure Blendit_MMX(bFr,bTo,bLn : Pointer; Width,Height : integer; Dens : LongInt);
asm
        push      edi
        push      esi
        db $0F,$77               /// emms
        mov       edi, Mask0101
        db $0F,$6E,$EF           /// movd      mm5, edi
        db $0F,$62,$ED           /// punpckldq mm5, mm5
        mov       edi, Mask1010
        db $0F,$6E,$E7           /// movd      mm4, edi
        db $0F,$62,$E4           /// punpckldq mm4, mm4
        mov       edi, Dens
        db $0F,$6E,$F7           /// movd      mm6, edi
        mov       edi, 64
        db $0F,$6E,$FF           /// movd      mm7, edi
        db $0F,$61,$FF           /// punpcklwd mm7, mm7
        db $0F,$62,$FF           /// punpckldq mm7, mm7
        db $0F,$61,$F6           /// punpcklwd mm6, mm6
        db $0F,$62,$F6           /// punpckldq mm6, mm6
        db $0F,$FA,$FE           /// psubd     mm7, mm6
        mov       edi, bFr
        mov       esi, bTo
        mov       ecx, bLn
        mov       eax, Width
        lea       eax, [eax+eax*2+3]
        and       eax, $FFFFFFFC
        imul      Height
        add       eax, edi
@LOOP:
        db $0F,$6F,$07           /// movq      mm0, [edi]
        db $0F,$6F,$0E           /// movq      mm1, [esi]
        db $0F,$6F,$D0           /// movq      mm2, mm0
        db $0F,$DB,$C4           /// pand      mm0, mm4
        db $0F,$DB,$D5           /// pand      mm2, mm5
        db $0F,$72,$D0,$06       /// psrld     mm0, 6
        db $0F,$6F,$D9           /// movq      mm3, mm1
        db $0F,$DB,$CC           /// pand      mm1, mm4
        db $0F,$DB,$DD           /// pand      mm3, mm5
        db $0F,$72,$D1,$06       /// psrld     mm1, 6
        db $0F,$D5,$C7           /// pmullw    mm0, mm7
        db $0F,$D5,$D7           /// pmullw    mm2, mm7
        db $0F,$D5,$CE           /// pmullw    mm1, mm6
        db $0F,$D5,$DE           /// pmullw    mm3, mm6
        db $0F,$FE,$C1           /// paddd     mm0, mm1
        db $0F,$FE,$D3           /// paddd     mm2, mm3
        db $0F,$DB,$C4           /// pand      mm0, mm4
        db $0F,$72,$D2,$06       /// psrld     mm2, 6
        db $0F,$DB,$D5           /// pand      mm2, mm5
        db $0F,$EB,$C2           /// por       mm0, mm2
        db $0F,$7F,$01           /// movq      [ecx], mm0
        add       edi, 8
        add       esi, 8
        add       ecx, 8
        cmp       edi, eax
        jl        @LOOP
@END:
        db $0F,$77               /// emms
        pop       edi
        pop       esi
end;

procedure Blendit(bFr,bTo,bLn : TBitmap; Dens : LongInt);
begin
  FBlend(bFr.Scanline[bFr.Height-1],bTo.Scanline[bTo.Height-1],
         bLn.Scanline[bLn.Height-1],bFr.Width,bFr.Height,Dens);
end;

var
  hasMMX: Boolean;

procedure CheckMMX;
asm
        mov       ecx, eax
        mov       hasMMX, 0
        pushfd
        pop       eax
        mov       ecx, eax
        xor       eax, $200000
        push      eax
        popfd
        pushfd
        pop       eax
        cmp       eax,ecx
        jz        @@END1
        pushad
        mov       eax, $1
        db $0F,$A2               /// CPUID
        test      edx, $800000
        jz        @@End2
        mov       hasMMX, 1
@@End2:
        popad
        xor       eax,eax
@@End1:
end;

initialization
  CheckMMX;
  if hasMMX
    then @FBlend := @BlendIt_MMX
    else @FBlend := @BlendIt_ASM;

end.
